@@ -39,7 +39,7 @@ typedef struct store_conf {
 	const char *flat_delim;
 	const char *map_inbox;
 	const char *trash;
-	uint max_size; /* off_t is overkill */
+	int max_size; /* off_t is overkill */
 	char trash_remote_new, trash_only_new;
 } store_conf_t;
 
@@ -63,9 +63,10 @@ typedef struct store_conf {
 typedef struct message {
 	struct message *next;
 	struct sync_rec *srec;
+	char *msgid; /* owned */
 	/* string_list_t *keywords; */
-	size_t size; /* zero implies "not fetched" */
-	int uid;
+	int size; /* zero implies "not fetched" */
+	uint uid;
 	uchar flags, status;
 	char tuid[TUIDL];
 } message_t;
@@ -74,41 +75,22 @@ typedef struct message {
 #define OPEN_OLD        (1<<0)
 #define OPEN_NEW        (1<<1)
 #define OPEN_FLAGS      (1<<2)
-#define OPEN_SIZE       (1<<3)
+#define OPEN_OLD_SIZE   (1<<3)
+#define OPEN_NEW_SIZE   (1<<4)
 #define OPEN_EXPUNGE    (1<<5)
 #define OPEN_SETFLAGS   (1<<6)
 #define OPEN_APPEND     (1<<7)
 #define OPEN_FIND       (1<<8)
+#define OPEN_OLD_IDS    (1<<9)
+
+#define UIDVAL_BAD ((uint)-1)
 
 typedef struct store {
 	struct store *next;
+	driver_t *driver;
 	store_conf_t *conf; /* foreign */
-	string_list_t *boxes; /* _list results - own */
-	char listed; /* was _list already run? */
-
-	void (*bad_callback)( void *aux );
-	void *bad_callback_aux;
-
-	/* currently open mailbox */
-	char *path; /* own */
-	message_t *msgs; /* own */
-	int uidvalidity;
-	int uidnext; /* from SELECT responses */
-	uint opts; /* maybe preset? */
-	/* note that the following do _not_ reflect stats from msgs, but mailbox totals */
-	int count; /* # of messages */
-	int recent; /* # of recent messages - don't trust this beyond the initial read */
 } store_t;
 
-/* When the callback is invoked (at most once per store), the store is fubar;
- * call the driver's cancel_store() to dispose of it. */
-static INLINE void
-set_bad_callback( store_t *ctx, void (*cb)( void *aux ), void *aux )
-{
-	ctx->bad_callback = cb;
-	ctx->bad_callback_aux = aux;
-}
-
 typedef struct {
 	char *data;
 	int len;
@@ -121,8 +103,10 @@ typedef struct {
 #define DRV_MSG_BAD     1
 /* Something is wrong with the current mailbox - probably it is somehow inaccessible. */
 #define DRV_BOX_BAD     2
+/* Failed to connect store. */
+#define DRV_STORE_BAD   3
 /* The command has been cancel()ed or cancel_store()d. */
-#define DRV_CANCELED    3
+#define DRV_CANCELED    4
 
 /* All memory belongs to the driver's user, unless stated otherwise. */
 
@@ -137,38 +121,52 @@ typedef struct {
 */
 #define DRV_VERBOSE     2
 
-#define LIST_PATH       1
-#define LIST_INBOX      2
+#define LIST_INBOX      1
+#define LIST_PATH       2
+#define LIST_PATH_MAYBE 4
+
+#define xint int  // For auto-generation of appropriate printf() formats.
 
 struct driver {
-	int flags;
+	/* Return driver capabilities. */
+	xint (*get_caps)( store_t *ctx );
 
 	/* Parse configuration. */
 	int (*parse_store)( conffile_t *cfg, store_conf_t **storep );
 
-	/* Close remaining server connections. All stores must be disowned first. */
+	/* Close remaining server connections. All stores must be discarded first. */
 	void (*cleanup)( void );
 
-	/* Open a store with the given configuration. This may recycle existing
-	 * server connections. Upon failure, a null store is passed to the callback. */
-	void (*open_store)( store_conf_t *conf, const char *label,
-	                    void (*cb)( store_t *ctx, void *aux ), void *aux );
+	/* Allocate a store with the given configuration. This is expected to
+	 * return quickly, and must not fail. */
+	store_t *(*alloc_store)( store_conf_t *conf, const char *label );
 
-	/* Mark the store as available for recycling. Server connection may be kept alive. */
-	void (*disown_store)( store_t *ctx );
+	/* When this callback is invoked (at most once per store), the store is fubar;
+	 * call cancel_store() to dispose of it. */
+	void (*set_bad_callback)( store_t *ctx, void (*cb)( void *aux ), void *aux );
+
+	/* Open/connect the store. This may recycle existing server connections. */
+	void (*connect_store)( store_t *ctx,
+	                       void (*cb)( int sts, void *aux ), void *aux );
+
+	/* Discard the store. Underlying server connection may be kept alive. */
+	void (*free_store)( store_t *ctx );
 
 	/* Discard the store after a bad_callback. The server connections will be closed.
 	 * Pending commands will have their callbacks synchronously invoked with DRV_CANCELED. */
 	void (*cancel_store)( store_t *ctx );
 
-	/* List the mailboxes in this store. Flags are ORed LIST_* values. */
+	/* List the mailboxes in this store. Flags are ORed LIST_* values.
+	 * The returned box list remains owned by the driver. */
 	void (*list_store)( store_t *ctx, int flags,
-	                    void (*cb)( int sts, void *aux ), void *aux );
+	                    void (*cb)( int sts, string_list_t *boxes, void *aux ), void *aux );
 
-	/* Invoked before open_box(), this informs the driver which box is to be opened.
-	 * As a side effect, this should resolve ctx->path if applicable. */
+	/* Invoked before open_box(), this informs the driver which box is to be opened. */
 	int (*select_box)( store_t *ctx, const char *name );
 
+	/* Get the selected box' on-disk path, if applicable, null otherwise. */
+	const char *(*get_box_path)( store_t *ctx );
+
 	/* Create the selected mailbox. */
 	void (*create_box)( store_t *ctx,
 	                    void (*cb)( int sts, void *aux ), void *aux );
@@ -176,7 +174,10 @@ struct driver {
 	/* Open the selected mailbox.
 	 * Note that this should not directly complain about failure to open. */
 	void (*open_box)( store_t *ctx,
-	                  void (*cb)( int sts, void *aux ), void *aux );
+	                  void (*cb)( int sts, int uidvalidity, void *aux ), void *aux );
+
+	/* Return the minimal UID the next stored message will have. */
+	int (*get_uidnext)( store_t *ctx );
 
 	/* Confirm that the open mailbox is empty. */
 	int (*confirm_box_empty)( store_t *ctx );
@@ -193,37 +194,42 @@ struct driver {
 
 	/* Invoked before load_box(), this informs the driver which operations (OP_*)
 	 * will be performed on the mailbox. The driver may extend the set by implicitly
-	 * needed or available operations. */
-	void (*prepare_load_box)( store_t *ctx, int opts );
+	 * needed or available operations. Returns this possibly extended set. */
+	xint (*prepare_load_box)( store_t *ctx, xint opts );
 
 	/* Load the message attributes needed to perform the requested operations.
 	 * Consider only messages with UIDs between minuid and maxuid (inclusive)
 	 * and those named in the excs array (smaller than minuid).
-	 * The driver takes ownership of the excs array. Messages below newuid do not need
-	 * to have the TUID populated even if OPEN_FIND is set. */
-	void (*load_box)( store_t *ctx, int minuid, int maxuid, int newuid, int *excs, int nexcs,
-	                  void (*cb)( int sts, void *aux ), void *aux );
+	 * The driver takes ownership of the excs array.
+	 * Messages starting with newuid need to have the TUID populated when OPEN_FIND is set.
+	 * Messages up to seenuid need to have the Message-Id populated when OPEN_OLD_IDS is set.
+	 * Messages up to seenuid need to have the size populated when OPEN_OLD_SIZE is set;
+	 * likewise messages above seenuid when OPEN_NEW_SIZE is set.
+	 * The returned message list remains owned by the driver. */
+	void (*load_box)( store_t *ctx, uint minuid, uint maxuid, uint newuid, uint seenuid, uint_array_t excs,
+	                  void (*cb)( int sts, message_t *msgs, int total_msgs, int recent_msgs, void *aux ), void *aux );
 
 	/* Fetch the contents and flags of the given message from the current mailbox. */
 	void (*fetch_msg)( store_t *ctx, message_t *msg, msg_data_t *data,
 	                   void (*cb)( int sts, void *aux ), void *aux );
 
 	/* Store the given message to either the current mailbox or the trash folder.
-	 * If the new copy's UID can be immediately determined, return it, otherwise -2. */
+	 * If the new copy's UID can be immediately determined, return it, otherwise 0. */
 	void (*store_msg)( store_t *ctx, msg_data_t *data, int to_trash,
-	                   void (*cb)( int sts, int uid, void *aux ), void *aux );
+	                   void (*cb)( int sts, uint uid, void *aux ), void *aux );
 
 	/* Index the messages which have newly appeared in the mailbox, including their
 	 * temporary UID headers. This is needed if store_msg() does not guarantee returning
-	 * a UID; otherwise the driver needs to implement only the OPEN_FIND flag. */
-	void (*find_new_msgs)( store_t *ctx, int newuid,
-	                       void (*cb)( int sts, void *aux ), void *aux );
+	 * a UID; otherwise the driver needs to implement only the OPEN_FIND flag.
+	 * The returned message list remains owned by the driver. */
+	void (*find_new_msgs)( store_t *ctx, uint newuid,
+	                       void (*cb)( int sts, message_t *msgs, void *aux ), void *aux );
 
 	/* Add/remove the named flags to/from the given message. The message may be either
 	 * a pre-fetched one (in which case the in-memory representation is updated),
 	 * or it may be identifed by UID only. The operation may be delayed until commit()
 	 * is called. */
-	void (*set_msg_flags)( store_t *ctx, message_t *msg, int uid, int add, int del, /* msg can be null, therefore uid as a fallback */
+	void (*set_msg_flags)( store_t *ctx, message_t *msg, uint uid, int add, int del, /* msg can be null, therefore uid as a fallback */
 	                       void (*cb)( int sts, void *aux ), void *aux );
 
 	/* Move the given message from the current mailbox to the trash folder.
@@ -247,18 +253,20 @@ struct driver {
 	void (*commit_cmds)( store_t *ctx );
 
 	/* Get approximate amount of memory occupied by the driver. */
-	int (*memory_usage)( store_t *ctx );
+	int (*get_memory_usage)( store_t *ctx );
 
 	/* Get the FAIL_* state of the driver. */
-	int (*fail_state)( store_conf_t *conf );
+	int (*get_fail_state)( store_conf_t *conf );
 };
 
 void free_generic_messages( message_t * );
 
 void parse_generic_store( store_conf_t *store, conffile_t *cfg );
 
+store_t *proxy_alloc_store( store_t *real_ctx, const char *label );
+
 #define N_DRIVERS 2
 extern driver_t *drivers[N_DRIVERS];
-extern driver_t maildir_driver, imap_driver;
+extern driver_t maildir_driver, imap_driver, proxy_driver;
 
 #endif
